/*
 * SPDX-FileCopyrightText: 2023 jordi fita mas <jfita@peritasoft.com>
 * SPDX-License-Identifier: AGPL-3.0-only
 */

package campsite

import (
	"context"
	"net/http"

	"dev.tandem.ws/tandem/camper/pkg/auth"
	"dev.tandem.ws/tandem/camper/pkg/campsite/types"
	"dev.tandem.ws/tandem/camper/pkg/database"
	"dev.tandem.ws/tandem/camper/pkg/form"
	httplib "dev.tandem.ws/tandem/camper/pkg/http"
	"dev.tandem.ws/tandem/camper/pkg/locale"
	"dev.tandem.ws/tandem/camper/pkg/template"
)

type AdminHandler struct {
	types *types.AdminHandler
}

func NewAdminHandler() *AdminHandler {
	return &AdminHandler{
		types: types.NewAdminHandler(),
	}
}

func (h *AdminHandler) Handler(user *auth.User, company *auth.Company, conn *database.Conn) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		var head string
		head, r.URL.Path = httplib.ShiftPath(r.URL.Path)

		switch head {
		case "new":
			switch r.Method {
			case http.MethodGet:
				f := newCampsiteForm(r.Context(), conn)
				f.MustRender(w, r, user, company)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet)
			}
		case "types":
			h.types.Handler(user, company, conn).ServeHTTP(w, r)
		case "":
			switch r.Method {
			case http.MethodGet:
				serveCampsiteIndex(w, r, user, company, conn)
			case http.MethodPost:
				addCampsite(w, r, user, company, conn)
			default:
				httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPost)
			}
		default:
			f := newCampsiteForm(r.Context(), conn)
			if err := f.FillFromDatabase(r.Context(), conn, company, head); err != nil {
				if database.ErrorIsNotFound(err) {
					http.NotFound(w, r)
					return
				}
				panic(err)
			}

			head, r.URL.Path = httplib.ShiftPath(r.URL.Path)
			switch head {
			case "":
				switch r.Method {
				case http.MethodGet:
					f.MustRender(w, r, user, company)
				case http.MethodPut:
					editCampsite(w, r, user, company, conn, f)
				default:
					httplib.MethodNotAllowed(w, r, http.MethodGet, http.MethodPut)
				}
			default:
				http.NotFound(w, r)
			}
		}
	}
}

func serveCampsiteIndex(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	campsites, err := collectCampsiteEntries(r.Context(), company, conn)
	if err != nil {
		panic(err)
	}
	page := &campsiteIndex{
		Campsites: campsites,
	}
	page.MustRender(w, r, user, company)
}

func collectCampsiteEntries(ctx context.Context, company *auth.Company, conn *database.Conn) ([]*campsiteEntry, error) {
	rows, err := conn.Query(ctx, `
		select campsite.label
		     , campsite_type.name
		     , campsite.active
		from campsite
		join campsite_type using (campsite_type_id)
		where campsite.company_id = $1
		order by label`, company.ID)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	var campsites []*campsiteEntry
	for rows.Next() {
		entry := &campsiteEntry{}
		if err = rows.Scan(&entry.Label, &entry.Type, &entry.Active); err != nil {
			return nil, err
		}
		campsites = append(campsites, entry)
	}

	return campsites, nil
}

type campsiteEntry struct {
	Label  string
	Type   string
	Active bool
}

type campsiteIndex struct {
	Campsites []*campsiteEntry
}

func (page *campsiteIndex) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdminFiles(w, r, user, company, page, "campsite/index.gohtml", "web/templates/campground_map.svg")
}

func addCampsite(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn) {
	f := newCampsiteForm(r.Context(), conn)
	processCampsiteForm(w, r, user, company, conn, f, func(ctx context.Context, tx *database.Tx) error {
		_, err := tx.AddCampsite(r.Context(), f.CampsiteType.String(), f.Label.Val)
		return err
	})
	httplib.Redirect(w, r, "/admin/campsites", http.StatusSeeOther)
}

func editCampsite(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *campsiteForm) {
	processCampsiteForm(w, r, user, company, conn, f, func(ctx context.Context, tx *database.Tx) error {
		return tx.EditCampsite(r.Context(), f.ID, f.CampsiteType.String(), f.Label.Val, f.Active.Checked)
	})
	httplib.Redirect(w, r, "/admin/campsites", http.StatusSeeOther)
}

func processCampsiteForm(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company, conn *database.Conn, f *campsiteForm, act func(ctx context.Context, tx *database.Tx) error) {
	if ok, err := form.Handle(f, w, r, user); err != nil {
		return
	} else if !ok {
		f.MustRender(w, r, user, company)
		return
	}

	tx := conn.MustBegin(r.Context())
	defer tx.Rollback(r.Context())
	if err := act(r.Context(), tx); err != nil {
		panic(err)
	}
	tx.MustCommit(r.Context())
	httplib.Redirect(w, r, "/admin/campsites", http.StatusSeeOther)
}

type campsiteForm struct {
	ID           int
	CurrentLabel string
	Active       *form.Checkbox
	CampsiteType *form.Select
	Label        *form.Input
}

func newCampsiteForm(ctx context.Context, conn *database.Conn) *campsiteForm {
	campsiteTypes := form.MustGetOptions(ctx, conn, "select campsite_type_id::text, name from campsite_type where active")
	return &campsiteForm{
		Active: &form.Checkbox{
			Name:    "active",
			Checked: true,
		},
		CampsiteType: &form.Select{
			Name:    "description",
			Options: campsiteTypes,
		},
		Label: &form.Input{
			Name: "label",
		},
	}
}

func (f *campsiteForm) FillFromDatabase(ctx context.Context, conn *database.Conn, company *auth.Company, label string) error {
	f.CurrentLabel = label
	row := conn.QueryRow(ctx, `
		select campsite_id
		     , array[campsite_type_id::text]
		     , label
		     , active
		from campsite
		where company_id = $1
		  and label = $2`, company.ID, label)
	return row.Scan(&f.ID, &f.CampsiteType.Selected, &f.Label.Val, &f.Active.Checked)
}

func (f *campsiteForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	f.Active.FillValue(r)
	f.CampsiteType.FillValue(r)
	f.Label.FillValue(r)
	return nil
}

func (f *campsiteForm) Valid(l *locale.Locale) bool {
	v := form.NewValidator(l)
	v.CheckSelectedOptions(f.CampsiteType, l.GettextNoop("Selected campsite type is not valid."))
	v.CheckRequired(f.Label, l.GettextNoop("Label can not be empty."))
	return v.AllOK
}

func (f *campsiteForm) MustRender(w http.ResponseWriter, r *http.Request, user *auth.User, company *auth.Company) {
	template.MustRenderAdmin(w, r, user, company, "campsite/form.gohtml", f)
}
